"""Perform queries against the variation effect predictor (VEP) endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Vep(object):
    """Perform Vep queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_vep_hgvs(self, hgvs_notation, species, blosum62=None, cadd=None,
                     conservation=None, gene_splicer=None, lo_f=None,
                     max_ent_scan=None, phenotypes=None, splice_ai=None,
                     splice_region=None, appris=None, canonical=None,
                     ccds=None, db_nsfp=None, dbsc_snv=None, distance=None,
                     domains=None, failed=None, hgvs=None, mane=None,
                     merged=None, mi_rna=None, minimal=None, numbers=None,
                     protein=None, refseq=None, shift_3prime=None,
                     shift_genomic=None, transcript_id=None,
                     transcript_version=None, tsl=None, uniprot=None,
                     variant_class=None, vcf_string=None, xref_refseq=None,
                     norm_func=brc.default_pandas):
        """
        Fetch variant consequences based on a HGVS notation. GET method.
        Queries:
        `vep/:species/hgvs/:hgvs_notation <https://rest.ensembl.org/documentation/info/vep_hgvs_get>`_.

        Parameters
        ----------
        hgvs_notation : str
            HGVS notation. May be genomic (g), coding (c) or protein (p), with
            reference to chromosome name, gene name, transcript ID or protein
            ID. Example values: 'AGT:c.803T>C 9:g.22125504G>C
            ENST00000003084:c.1431_1433delTTC'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        blosum62 : bool or NoneType, optional, default: NoneType
            Include BLOSUM62 amino acid conservation score (plugin details).
        cadd : bool or NoneType, optional, default: NoneType
            Include CADD (Combined Annotation Dependent Depletion)
            deleteriousness scores for single nucleotide variants. See license.
            (plugin details).
        conservation : bool or NoneType, optional, default: NoneType
            Retrieves a conservation score from the Ensembl Compara databases
            for variant positions (plugin details).
        gene_splicer : bool or NoneType, optional, default: NoneType
            Detects splice sites in genomic DNA (plugin details).
        lo_f : bool or NoneType, optional, default: NoneType
            LOFTEE identifies LoF (loss-of-function) variation. See README for
            more details.
        max_ent_scan : bool or NoneType, optional, default: NoneType
            Sequence motif and maximum entropy based splice site consensus
            predictions (plugin details).
        phenotypes : bool or NoneType, optional, default: NoneType
            Retrieves overlapping phenotype information (plugin details).
        splice_ai : bool or NoneType, optional, default: NoneType
            Retrieves pre-calculated annotations from SpliceAI a deep neural
            network, developed by Illumina, Inc that predicts splice junctions
            from an arbitrary pre-mRNA transcript sequence. Used for non-
            commercial purposes. (plugin details).
        splice_region : bool or NoneType, optional, default: NoneType
            More granular predictions of splicing effects (plugin details).
        appris : bool or NoneType, optional, default: NoneType
            Include APPRIS isoform annotation.
        canonical : bool or NoneType, optional, default: NoneType
            Include a flag indicating the canonical transcript for a gene.
        ccds : bool or NoneType, optional, default: NoneType
            Include CCDS transcript identifiers.
        db_nsfp : str or NoneType, optional, default: NoneType
            Include fields from dbNSFP, a database of pathogenicity predictions
            for missense variants. Multiple fields should be separated by
            commas. See dbNSFP README for field list. (plugin details). Example
            values: 'LRT_pred,MutationTaster_pred'.
        dbsc_snv : bool or NoneType, optional, default: NoneType
            Predictions for splicing variants from dbscSNV. (plugin details).
        distance : int or NoneType, optional, default: NoneType
            Change the distance to transcript for which VEP assigns upstream
            and downstream consequences.
        domains : bool or NoneType, optional, default: NoneType
            Include names of overlapping protein domains.
        failed : bool or NoneType, optional, default: NoneType
            When checking for co-located variants, by default variants flagged
            as failed by Ensembl's QC pipeline will be excluded. Set this flag
            to 1 to include such variants.
        hgvs : bool or NoneType, optional, default: NoneType
            Include HGVS nomenclature based on Ensembl stable identifiers.
        mane : bool or NoneType, optional, default: NoneType
            Include MANE Select annotations (GRCh38 only).
        merged : bool or NoneType, optional, default: NoneType
            Use merged Ensembl and RefSeq transcript set to report consequences
            (human only).
        mi_rna : bool or NoneType, optional, default: NoneType
            Determines where in the secondary structure of a miRNA a variant
            falls (plugin details).
        minimal : bool or NoneType, optional, default: NoneType
            Convert alleles to their most minimal representation before
            consequence calculation i.e. sequence that is identical between
            each pair of reference and alternate alleles is trimmed off from
            both ends, with coordinates adjusted accordingly. Note this may
            lead to discrepancies between input coordinates and coordinates
            reported by VEP relative to transcript sequences.
        numbers : bool or NoneType, optional, default: NoneType
            Include affected exon and intron positions within the transcript.
        protein : bool or NoneType, optional, default: NoneType
            Include Ensembl protein identifiers.
        refseq : bool or NoneType, optional, default: NoneType
            Use RefSeq transcript set to report consequences (human only).
        shift_3prime : bool or NoneType, optional, default: NoneType
            Shift transcript-overlapping variants as far as possible in the 3'
            direction before providing consequences.
        shift_genomic : bool or NoneType, optional, default: NoneType
            Shift all variants as far as possible in the 3' direction before
            providing consequences.
        transcript_id : str or NoneType, optional, default: NoneType
            Filter results by Transcript ID.
        transcript_version : bool or NoneType, optional, default: NoneType
            Add version numbers to Ensembl transcript identifiers.
        tsl : bool or NoneType, optional, default: NoneType
            Include transcript support level (TSL) annotation.
        uniprot : bool or NoneType, optional, default: NoneType
            Include best match accessions for translated protein products from
            three UniProt-related databases (SWISSPROT, TREMBL and UniParc).
        variant_class : bool or NoneType, optional, default: NoneType
            Output the Sequence Ontology variant class for the input variant.
        vcf_string : bool or NoneType, optional, default: NoneType
            Include alleles in VCF format.
        xref_refseq : bool or NoneType, optional, default: NoneType
            Include aligned RefSeq mRNA identifiers for transcript. NB:
            theRefSeq and Ensembl transcripts aligned in this way MAY NOT, AND
            FREQUENTLY WILL NOT, match exactly in sequence, exon structure and
            protein product.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'vep/{species}/hgvs/{hgvs_notation}'.format(
            species=species,
            hgvs_notation=hgvs_notation
        )

        return self.rest_query(
            method_type,
            endpoint,
            Blosum62=blosum62,
            CADD=cadd,
            Conservation=conservation,
            GeneSplicer=gene_splicer,
            LoF=lo_f,
            MaxEntScan=max_ent_scan,
            Phenotypes=phenotypes,
            SpliceAI=splice_ai,
            SpliceRegion=splice_region,
            appris=appris,
            canonical=canonical,
            ccds=ccds,
            dbNSFP=db_nsfp,
            dbscSNV=dbsc_snv,
            distance=distance,
            domains=domains,
            failed=failed,
            hgvs=hgvs,
            mane=mane,
            merged=merged,
            miRNA=mi_rna,
            minimal=minimal,
            numbers=numbers,
            protein=protein,
            refseq=refseq,
            shift_3prime=shift_3prime,
            shift_genomic=shift_genomic,
            transcript_id=transcript_id,
            transcript_version=transcript_version,
            tsl=tsl,
            uniprot=uniprot,
            variant_class=variant_class,
            vcf_string=vcf_string,
            xref_refseq=xref_refseq,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_vep_hgvs(self, species, hgvs_notations, blosum62=None, cadd=None,
                      gene_splicer=None, lo_f=None, max_ent_scan=None,
                      phenotypes=None, splice_ai=None, splice_region=None,
                      appris=None, canonical=None, ccds=None, db_nsfp=None,
                      dbsc_snv=None, distance=None, domains=None, failed=None,
                      hgvs=None, mane=None, merged=None, mi_rna=None,
                      minimal=None, numbers=None, protein=None, refseq=None,
                      shift_3prime=None, shift_genomic=None,
                      transcript_id=None, transcript_version=None, tsl=None,
                      uniprot=None, variant_class=None, vcf_string=None,
                      xref_refseq=None, norm_func=brc.default_pandas):
        """
        Fetch variant consequences for multiple HGVS notations. POST method.
        Queries:
        `vep/:species/hgvs <https://rest.ensembl.org/documentation/info/vep_hgvs_post>`_.
        Max post size: 200.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        hgvs_notations : list
            Post option not defined in required/optional args.
        blosum62 : bool or NoneType, optional, default: NoneType
            Include BLOSUM62 amino acid conservation score (plugin details).
        cadd : bool or NoneType, optional, default: NoneType
            Include CADD (Combined Annotation Dependent Depletion)
            deleteriousness scores for single nucleotide variants. See license.
            (plugin details).
        gene_splicer : bool or NoneType, optional, default: NoneType
            Detects splice sites in genomic DNA (plugin details).
        lo_f : bool or NoneType, optional, default: NoneType
            LOFTEE identifies LoF (loss-of-function) variation. See README for
            more details.
        max_ent_scan : bool or NoneType, optional, default: NoneType
            Sequence motif and maximum entropy based splice site consensus
            predictions (plugin details).
        phenotypes : bool or NoneType, optional, default: NoneType
            Retrieves overlapping phenotype information (plugin details).
        splice_ai : bool or NoneType, optional, default: NoneType
            Retrieves pre-calculated annotations from SpliceAI a deep neural
            network, developed by Illumina, Inc that predicts splice junctions
            from an arbitrary pre-mRNA transcript sequence. Used for non-
            commercial purposes. (plugin details).
        splice_region : bool or NoneType, optional, default: NoneType
            More granular predictions of splicing effects (plugin details).
        appris : bool or NoneType, optional, default: NoneType
            Include APPRIS isoform annotation.
        canonical : bool or NoneType, optional, default: NoneType
            Include a flag indicating the canonical transcript for a gene.
        ccds : bool or NoneType, optional, default: NoneType
            Include CCDS transcript identifiers.
        db_nsfp : str or NoneType, optional, default: NoneType
            Include fields from dbNSFP, a database of pathogenicity predictions
            for missense variants. Multiple fields should be separated by
            commas. See dbNSFP README for field list. (plugin details). Example
            values: 'LRT_pred,MutationTaster_pred'.
        dbsc_snv : bool or NoneType, optional, default: NoneType
            Predictions for splicing variants from dbscSNV. (plugin details).
        distance : int or NoneType, optional, default: NoneType
            Change the distance to transcript for which VEP assigns upstream
            and downstream consequences.
        domains : bool or NoneType, optional, default: NoneType
            Include names of overlapping protein domains.
        failed : bool or NoneType, optional, default: NoneType
            When checking for co-located variants, by default variants flagged
            as failed by Ensembl's QC pipeline will be excluded. Set this flag
            to 1 to include such variants.
        hgvs : bool or NoneType, optional, default: NoneType
            Include HGVS nomenclature based on Ensembl stable identifiers.
        mane : bool or NoneType, optional, default: NoneType
            Include MANE Select annotations (GRCh38 only).
        merged : bool or NoneType, optional, default: NoneType
            Use merged Ensembl and RefSeq transcript set to report consequences
            (human only).
        mi_rna : bool or NoneType, optional, default: NoneType
            Determines where in the secondary structure of a miRNA a variant
            falls (plugin details).
        minimal : bool or NoneType, optional, default: NoneType
            Convert alleles to their most minimal representation before
            consequence calculation i.e. sequence that is identical between
            each pair of reference and alternate alleles is trimmed off from
            both ends, with coordinates adjusted accordingly. Note this may
            lead to discrepancies between input coordinates and coordinates
            reported by VEP relative to transcript sequences.
        numbers : bool or NoneType, optional, default: NoneType
            Include affected exon and intron positions within the transcript.
        protein : bool or NoneType, optional, default: NoneType
            Include Ensembl protein identifiers.
        refseq : bool or NoneType, optional, default: NoneType
            Use RefSeq transcript set to report consequences (human only).
        shift_3prime : bool or NoneType, optional, default: NoneType
            Shift transcript-overlapping variants as far as possible in the 3'
            direction before providing consequences.
        shift_genomic : bool or NoneType, optional, default: NoneType
            Shift all variants as far as possible in the 3' direction before
            providing consequences.
        transcript_id : str or NoneType, optional, default: NoneType
            Filter results by Transcript ID.
        transcript_version : bool or NoneType, optional, default: NoneType
            Add version numbers to Ensembl transcript identifiers.
        tsl : bool or NoneType, optional, default: NoneType
            Include transcript support level (TSL) annotation.
        uniprot : bool or NoneType, optional, default: NoneType
            Include best match accessions for translated protein products from
            three UniProt-related databases (SWISSPROT, TREMBL and UniParc).
        variant_class : bool or NoneType, optional, default: NoneType
            Output the Sequence Ontology variant class for the input variant.
        vcf_string : bool or NoneType, optional, default: NoneType
            Include alleles in VCF format.
        xref_refseq : bool or NoneType, optional, default: NoneType
            Include aligned RefSeq mRNA identifiers for transcript. NB:
            theRefSeq and Ensembl transcripts aligned in this way MAY NOT, AND
            FREQUENTLY WILL NOT, match exactly in sequence, exon structure and
            protein product.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        maximum_post_size = 200
        endpoint = 'vep/{species}/hgvs'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            hgvs_notations=hgvs_notations,
            Blosum62=blosum62,
            CADD=cadd,
            GeneSplicer=gene_splicer,
            LoF=lo_f,
            MaxEntScan=max_ent_scan,
            Phenotypes=phenotypes,
            SpliceAI=splice_ai,
            SpliceRegion=splice_region,
            appris=appris,
            canonical=canonical,
            ccds=ccds,
            dbNSFP=db_nsfp,
            dbscSNV=dbsc_snv,
            distance=distance,
            domains=domains,
            failed=failed,
            hgvs=hgvs,
            mane=mane,
            merged=merged,
            miRNA=mi_rna,
            minimal=minimal,
            numbers=numbers,
            protein=protein,
            refseq=refseq,
            shift_3prime=shift_3prime,
            shift_genomic=shift_genomic,
            transcript_id=transcript_id,
            transcript_version=transcript_version,
            tsl=tsl,
            uniprot=uniprot,
            variant_class=variant_class,
            vcf_string=vcf_string,
            xref_refseq=xref_refseq,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_vep_id(self, query_id, species, blosum62=None, cadd=None,
                   conservation=None, gene_splicer=None, lo_f=None,
                   max_ent_scan=None, phenotypes=None, splice_ai=None,
                   splice_region=None, appris=None, canonical=None, ccds=None,
                   db_nsfp=None, dbsc_snv=None, distance=None, domains=None,
                   failed=None, hgvs=None, mane=None, merged=None, mi_rna=None,
                   minimal=None, numbers=None, protein=None, refseq=None,
                   shift_3prime=None, shift_genomic=None, transcript_id=None,
                   transcript_version=None, tsl=None, uniprot=None,
                   variant_class=None, vcf_string=None, xref_refseq=None,
                   norm_func=brc.default_pandas):
        """
        Fetch variant consequences based on a variant identifier. GET method.
        Queries:
        `vep/:species/id/:id <https://rest.ensembl.org/documentation/info/vep_id_get>`_.

        Parameters
        ----------
        query_id : str
            Query ID. Supports dbSNP, COSMIC and HGMD identifiers. Example
            values: 'rs56116432 COSM476'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        blosum62 : bool or NoneType, optional, default: NoneType
            Include BLOSUM62 amino acid conservation score (plugin details).
        cadd : bool or NoneType, optional, default: NoneType
            Include CADD (Combined Annotation Dependent Depletion)
            deleteriousness scores for single nucleotide variants. See license.
            (plugin details).
        conservation : bool or NoneType, optional, default: NoneType
            Retrieves a conservation score from the Ensembl Compara databases
            for variant positions (plugin details).
        gene_splicer : bool or NoneType, optional, default: NoneType
            Detects splice sites in genomic DNA (plugin details).
        lo_f : bool or NoneType, optional, default: NoneType
            LOFTEE identifies LoF (loss-of-function) variation. See README for
            more details.
        max_ent_scan : bool or NoneType, optional, default: NoneType
            Sequence motif and maximum entropy based splice site consensus
            predictions (plugin details).
        phenotypes : bool or NoneType, optional, default: NoneType
            Retrieves overlapping phenotype information (plugin details).
        splice_ai : bool or NoneType, optional, default: NoneType
            Retrieves pre-calculated annotations from SpliceAI a deep neural
            network, developed by Illumina, Inc that predicts splice junctions
            from an arbitrary pre-mRNA transcript sequence. Used for non-
            commercial purposes. (plugin details).
        splice_region : bool or NoneType, optional, default: NoneType
            More granular predictions of splicing effects (plugin details).
        appris : bool or NoneType, optional, default: NoneType
            Include APPRIS isoform annotation.
        canonical : bool or NoneType, optional, default: NoneType
            Include a flag indicating the canonical transcript for a gene.
        ccds : bool or NoneType, optional, default: NoneType
            Include CCDS transcript identifiers.
        db_nsfp : str or NoneType, optional, default: NoneType
            Include fields from dbNSFP, a database of pathogenicity predictions
            for missense variants. Multiple fields should be separated by
            commas. See dbNSFP README for field list. (plugin details). Example
            values: 'LRT_pred,MutationTaster_pred'.
        dbsc_snv : bool or NoneType, optional, default: NoneType
            Predictions for splicing variants from dbscSNV. (plugin details).
        distance : int or NoneType, optional, default: NoneType
            Change the distance to transcript for which VEP assigns upstream
            and downstream consequences.
        domains : bool or NoneType, optional, default: NoneType
            Include names of overlapping protein domains.
        failed : bool or NoneType, optional, default: NoneType
            When checking for co-located variants, by default variants flagged
            as failed by Ensembl's QC pipeline will be excluded. Set this flag
            to 1 to include such variants.
        hgvs : bool or NoneType, optional, default: NoneType
            Include HGVS nomenclature based on Ensembl stable identifiers.
        mane : bool or NoneType, optional, default: NoneType
            Include MANE Select annotations (GRCh38 only).
        merged : bool or NoneType, optional, default: NoneType
            Use merged Ensembl and RefSeq transcript set to report consequences
            (human only).
        mi_rna : bool or NoneType, optional, default: NoneType
            Determines where in the secondary structure of a miRNA a variant
            falls (plugin details).
        minimal : bool or NoneType, optional, default: NoneType
            Convert alleles to their most minimal representation before
            consequence calculation i.e. sequence that is identical between
            each pair of reference and alternate alleles is trimmed off from
            both ends, with coordinates adjusted accordingly. Note this may
            lead to discrepancies between input coordinates and coordinates
            reported by VEP relative to transcript sequences.
        numbers : bool or NoneType, optional, default: NoneType
            Include affected exon and intron positions within the transcript.
        protein : bool or NoneType, optional, default: NoneType
            Include Ensembl protein identifiers.
        refseq : bool or NoneType, optional, default: NoneType
            Use RefSeq transcript set to report consequences (human only).
        shift_3prime : bool or NoneType, optional, default: NoneType
            Shift transcript-overlapping variants as far as possible in the 3'
            direction before providing consequences.
        shift_genomic : bool or NoneType, optional, default: NoneType
            Shift all variants as far as possible in the 3' direction before
            providing consequences.
        transcript_id : str or NoneType, optional, default: NoneType
            Filter results by Transcript ID.
        transcript_version : bool or NoneType, optional, default: NoneType
            Add version numbers to Ensembl transcript identifiers.
        tsl : bool or NoneType, optional, default: NoneType
            Include transcript support level (TSL) annotation.
        uniprot : bool or NoneType, optional, default: NoneType
            Include best match accessions for translated protein products from
            three UniProt-related databases (SWISSPROT, TREMBL and UniParc).
        variant_class : bool or NoneType, optional, default: NoneType
            Output the Sequence Ontology variant class for the input variant.
        vcf_string : bool or NoneType, optional, default: NoneType
            Include alleles in VCF format.
        xref_refseq : bool or NoneType, optional, default: NoneType
            Include aligned RefSeq mRNA identifiers for transcript. NB:
            theRefSeq and Ensembl transcripts aligned in this way MAY NOT, AND
            FREQUENTLY WILL NOT, match exactly in sequence, exon structure and
            protein product.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'vep/{species}/id/{id}'.format(
            species=species,
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            Blosum62=blosum62,
            CADD=cadd,
            Conservation=conservation,
            GeneSplicer=gene_splicer,
            LoF=lo_f,
            MaxEntScan=max_ent_scan,
            Phenotypes=phenotypes,
            SpliceAI=splice_ai,
            SpliceRegion=splice_region,
            appris=appris,
            canonical=canonical,
            ccds=ccds,
            dbNSFP=db_nsfp,
            dbscSNV=dbsc_snv,
            distance=distance,
            domains=domains,
            failed=failed,
            hgvs=hgvs,
            mane=mane,
            merged=merged,
            miRNA=mi_rna,
            minimal=minimal,
            numbers=numbers,
            protein=protein,
            refseq=refseq,
            shift_3prime=shift_3prime,
            shift_genomic=shift_genomic,
            transcript_id=transcript_id,
            transcript_version=transcript_version,
            tsl=tsl,
            uniprot=uniprot,
            variant_class=variant_class,
            vcf_string=vcf_string,
            xref_refseq=xref_refseq,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_vep_id(self, species, ids, blosum62=None, cadd=None,
                    gene_splicer=None, lo_f=None, max_ent_scan=None,
                    phenotypes=None, splice_ai=None, splice_region=None,
                    appris=None, canonical=None, ccds=None, db_nsfp=None,
                    dbsc_snv=None, distance=None, domains=None, failed=None,
                    hgvs=None, mane=None, merged=None, mi_rna=None,
                    minimal=None, numbers=None, protein=None, refseq=None,
                    shift_3prime=None, shift_genomic=None, transcript_id=None,
                    transcript_version=None, tsl=None, uniprot=None,
                    variant_class=None, vcf_string=None, xref_refseq=None,
                    norm_func=brc.default_pandas):
        """
        Fetch variant consequences for multiple ids. POST method. Queries:
        `vep/:species/id <https://rest.ensembl.org/documentation/info/vep_id_post>`_.
        Max post size: 200.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        ids : list
            Post option not defined in required/optional args.
        blosum62 : bool or NoneType, optional, default: NoneType
            Include BLOSUM62 amino acid conservation score (plugin details).
        cadd : bool or NoneType, optional, default: NoneType
            Include CADD (Combined Annotation Dependent Depletion)
            deleteriousness scores for single nucleotide variants. See license.
            (plugin details).
        gene_splicer : bool or NoneType, optional, default: NoneType
            Detects splice sites in genomic DNA (plugin details).
        lo_f : bool or NoneType, optional, default: NoneType
            LOFTEE identifies LoF (loss-of-function) variation. See README for
            more details.
        max_ent_scan : bool or NoneType, optional, default: NoneType
            Sequence motif and maximum entropy based splice site consensus
            predictions (plugin details).
        phenotypes : bool or NoneType, optional, default: NoneType
            Retrieves overlapping phenotype information (plugin details).
        splice_ai : bool or NoneType, optional, default: NoneType
            Retrieves pre-calculated annotations from SpliceAI a deep neural
            network, developed by Illumina, Inc that predicts splice junctions
            from an arbitrary pre-mRNA transcript sequence. Used for non-
            commercial purposes. (plugin details).
        splice_region : bool or NoneType, optional, default: NoneType
            More granular predictions of splicing effects (plugin details).
        appris : bool or NoneType, optional, default: NoneType
            Include APPRIS isoform annotation.
        canonical : bool or NoneType, optional, default: NoneType
            Include a flag indicating the canonical transcript for a gene.
        ccds : bool or NoneType, optional, default: NoneType
            Include CCDS transcript identifiers.
        db_nsfp : str or NoneType, optional, default: NoneType
            Include fields from dbNSFP, a database of pathogenicity predictions
            for missense variants. Multiple fields should be separated by
            commas. See dbNSFP README for field list. (plugin details). Example
            values: 'LRT_pred,MutationTaster_pred'.
        dbsc_snv : bool or NoneType, optional, default: NoneType
            Predictions for splicing variants from dbscSNV. (plugin details).
        distance : int or NoneType, optional, default: NoneType
            Change the distance to transcript for which VEP assigns upstream
            and downstream consequences.
        domains : bool or NoneType, optional, default: NoneType
            Include names of overlapping protein domains.
        failed : bool or NoneType, optional, default: NoneType
            When checking for co-located variants, by default variants flagged
            as failed by Ensembl's QC pipeline will be excluded. Set this flag
            to 1 to include such variants.
        hgvs : bool or NoneType, optional, default: NoneType
            Include HGVS nomenclature based on Ensembl stable identifiers.
        mane : bool or NoneType, optional, default: NoneType
            Include MANE Select annotations (GRCh38 only).
        merged : bool or NoneType, optional, default: NoneType
            Use merged Ensembl and RefSeq transcript set to report consequences
            (human only).
        mi_rna : bool or NoneType, optional, default: NoneType
            Determines where in the secondary structure of a miRNA a variant
            falls (plugin details).
        minimal : bool or NoneType, optional, default: NoneType
            Convert alleles to their most minimal representation before
            consequence calculation i.e. sequence that is identical between
            each pair of reference and alternate alleles is trimmed off from
            both ends, with coordinates adjusted accordingly. Note this may
            lead to discrepancies between input coordinates and coordinates
            reported by VEP relative to transcript sequences.
        numbers : bool or NoneType, optional, default: NoneType
            Include affected exon and intron positions within the transcript.
        protein : bool or NoneType, optional, default: NoneType
            Include Ensembl protein identifiers.
        refseq : bool or NoneType, optional, default: NoneType
            Use RefSeq transcript set to report consequences (human only).
        shift_3prime : bool or NoneType, optional, default: NoneType
            Shift transcript-overlapping variants as far as possible in the 3'
            direction before providing consequences.
        shift_genomic : bool or NoneType, optional, default: NoneType
            Shift all variants as far as possible in the 3' direction before
            providing consequences.
        transcript_id : str or NoneType, optional, default: NoneType
            Filter results by Transcript ID.
        transcript_version : bool or NoneType, optional, default: NoneType
            Add version numbers to Ensembl transcript identifiers.
        tsl : bool or NoneType, optional, default: NoneType
            Include transcript support level (TSL) annotation.
        uniprot : bool or NoneType, optional, default: NoneType
            Include best match accessions for translated protein products from
            three UniProt-related databases (SWISSPROT, TREMBL and UniParc).
        variant_class : bool or NoneType, optional, default: NoneType
            Output the Sequence Ontology variant class for the input variant.
        vcf_string : bool or NoneType, optional, default: NoneType
            Include alleles in VCF format.
        xref_refseq : bool or NoneType, optional, default: NoneType
            Include aligned RefSeq mRNA identifiers for transcript. NB:
            theRefSeq and Ensembl transcripts aligned in this way MAY NOT, AND
            FREQUENTLY WILL NOT, match exactly in sequence, exon structure and
            protein product.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        maximum_post_size = 200
        endpoint = 'vep/{species}/id'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            ids=ids,
            Blosum62=blosum62,
            CADD=cadd,
            GeneSplicer=gene_splicer,
            LoF=lo_f,
            MaxEntScan=max_ent_scan,
            Phenotypes=phenotypes,
            SpliceAI=splice_ai,
            SpliceRegion=splice_region,
            appris=appris,
            canonical=canonical,
            ccds=ccds,
            dbNSFP=db_nsfp,
            dbscSNV=dbsc_snv,
            distance=distance,
            domains=domains,
            failed=failed,
            hgvs=hgvs,
            mane=mane,
            merged=merged,
            miRNA=mi_rna,
            minimal=minimal,
            numbers=numbers,
            protein=protein,
            refseq=refseq,
            shift_3prime=shift_3prime,
            shift_genomic=shift_genomic,
            transcript_id=transcript_id,
            transcript_version=transcript_version,
            tsl=tsl,
            uniprot=uniprot,
            variant_class=variant_class,
            vcf_string=vcf_string,
            xref_refseq=xref_refseq,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_vep_region(self, allele, region, species, blosum62=None, cadd=None,
                       conservation=None, gene_splicer=None, lo_f=None,
                       max_ent_scan=None, phenotypes=None, splice_ai=None,
                       splice_region=None, appris=None, canonical=None,
                       ccds=None, db_nsfp=None, dbsc_snv=None, distance=None,
                       domains=None, failed=None, hgvs=None, mane=None,
                       merged=None, mi_rna=None, minimal=None, numbers=None,
                       protein=None, refseq=None, shift_3prime=None,
                       shift_genomic=None, transcript_id=None,
                       transcript_version=None, tsl=None, uniprot=None,
                       variant_class=None, vcf_string=None, xref_refseq=None,
                       norm_func=brc.default_pandas):
        """
        Fetch variant consequences. GET method. Queries:
        `vep/:species/region/:region/:allele/ <https://rest.ensembl.org/documentation/info/vep_region_get>`_.

        Parameters
        ----------
        allele : str
            Variation allele. Example values: 'C DUP'.
        region : str
            Query region. We only support the current assembly. Example values:
            '9:22125503-22125502:1 7:100318423-100321323:1'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        blosum62 : bool or NoneType, optional, default: NoneType
            Include BLOSUM62 amino acid conservation score (plugin details).
        cadd : bool or NoneType, optional, default: NoneType
            Include CADD (Combined Annotation Dependent Depletion)
            deleteriousness scores for single nucleotide variants. See license.
            (plugin details).
        conservation : bool or NoneType, optional, default: NoneType
            Retrieves a conservation score from the Ensembl Compara databases
            for variant positions (plugin details).
        gene_splicer : bool or NoneType, optional, default: NoneType
            Detects splice sites in genomic DNA (plugin details).
        lo_f : bool or NoneType, optional, default: NoneType
            LOFTEE identifies LoF (loss-of-function) variation. See README for
            more details.
        max_ent_scan : bool or NoneType, optional, default: NoneType
            Sequence motif and maximum entropy based splice site consensus
            predictions (plugin details).
        phenotypes : bool or NoneType, optional, default: NoneType
            Retrieves overlapping phenotype information (plugin details).
        splice_ai : bool or NoneType, optional, default: NoneType
            Retrieves pre-calculated annotations from SpliceAI a deep neural
            network, developed by Illumina, Inc that predicts splice junctions
            from an arbitrary pre-mRNA transcript sequence. Used for non-
            commercial purposes. (plugin details).
        splice_region : bool or NoneType, optional, default: NoneType
            More granular predictions of splicing effects (plugin details).
        appris : bool or NoneType, optional, default: NoneType
            Include APPRIS isoform annotation.
        canonical : bool or NoneType, optional, default: NoneType
            Include a flag indicating the canonical transcript for a gene.
        ccds : bool or NoneType, optional, default: NoneType
            Include CCDS transcript identifiers.
        db_nsfp : str or NoneType, optional, default: NoneType
            Include fields from dbNSFP, a database of pathogenicity predictions
            for missense variants. Multiple fields should be separated by
            commas. See dbNSFP README for field list. (plugin details). Example
            values: 'LRT_pred,MutationTaster_pred'.
        dbsc_snv : bool or NoneType, optional, default: NoneType
            Predictions for splicing variants from dbscSNV. (plugin details).
        distance : int or NoneType, optional, default: NoneType
            Change the distance to transcript for which VEP assigns upstream
            and downstream consequences.
        domains : bool or NoneType, optional, default: NoneType
            Include names of overlapping protein domains.
        failed : bool or NoneType, optional, default: NoneType
            When checking for co-located variants, by default variants flagged
            as failed by Ensembl's QC pipeline will be excluded. Set this flag
            to 1 to include such variants.
        hgvs : bool or NoneType, optional, default: NoneType
            Include HGVS nomenclature based on Ensembl stable identifiers.
        mane : bool or NoneType, optional, default: NoneType
            Include MANE Select annotations (GRCh38 only).
        merged : bool or NoneType, optional, default: NoneType
            Use merged Ensembl and RefSeq transcript set to report consequences
            (human only).
        mi_rna : bool or NoneType, optional, default: NoneType
            Determines where in the secondary structure of a miRNA a variant
            falls (plugin details).
        minimal : bool or NoneType, optional, default: NoneType
            Convert alleles to their most minimal representation before
            consequence calculation i.e. sequence that is identical between
            each pair of reference and alternate alleles is trimmed off from
            both ends, with coordinates adjusted accordingly. Note this may
            lead to discrepancies between input coordinates and coordinates
            reported by VEP relative to transcript sequences.
        numbers : bool or NoneType, optional, default: NoneType
            Include affected exon and intron positions within the transcript.
        protein : bool or NoneType, optional, default: NoneType
            Include Ensembl protein identifiers.
        refseq : bool or NoneType, optional, default: NoneType
            Use RefSeq transcript set to report consequences (human only).
        shift_3prime : bool or NoneType, optional, default: NoneType
            Shift transcript-overlapping variants as far as possible in the 3'
            direction before providing consequences.
        shift_genomic : bool or NoneType, optional, default: NoneType
            Shift all variants as far as possible in the 3' direction before
            providing consequences.
        transcript_id : str or NoneType, optional, default: NoneType
            Filter results by Transcript ID.
        transcript_version : bool or NoneType, optional, default: NoneType
            Add version numbers to Ensembl transcript identifiers.
        tsl : bool or NoneType, optional, default: NoneType
            Include transcript support level (TSL) annotation.
        uniprot : bool or NoneType, optional, default: NoneType
            Include best match accessions for translated protein products from
            three UniProt-related databases (SWISSPROT, TREMBL and UniParc).
        variant_class : bool or NoneType, optional, default: NoneType
            Output the Sequence Ontology variant class for the input variant.
        vcf_string : bool or NoneType, optional, default: NoneType
            Include alleles in VCF format.
        xref_refseq : bool or NoneType, optional, default: NoneType
            Include aligned RefSeq mRNA identifiers for transcript. NB:
            theRefSeq and Ensembl transcripts aligned in this way MAY NOT, AND
            FREQUENTLY WILL NOT, match exactly in sequence, exon structure and
            protein product.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'vep/{species}/region/{region}/{allele}/'.format(
            species=species,
            region=region,
            allele=allele
        )

        return self.rest_query(
            method_type,
            endpoint,
            Blosum62=blosum62,
            CADD=cadd,
            Conservation=conservation,
            GeneSplicer=gene_splicer,
            LoF=lo_f,
            MaxEntScan=max_ent_scan,
            Phenotypes=phenotypes,
            SpliceAI=splice_ai,
            SpliceRegion=splice_region,
            appris=appris,
            canonical=canonical,
            ccds=ccds,
            dbNSFP=db_nsfp,
            dbscSNV=dbsc_snv,
            distance=distance,
            domains=domains,
            failed=failed,
            hgvs=hgvs,
            mane=mane,
            merged=merged,
            miRNA=mi_rna,
            minimal=minimal,
            numbers=numbers,
            protein=protein,
            refseq=refseq,
            shift_3prime=shift_3prime,
            shift_genomic=shift_genomic,
            transcript_id=transcript_id,
            transcript_version=transcript_version,
            tsl=tsl,
            uniprot=uniprot,
            variant_class=variant_class,
            vcf_string=vcf_string,
            xref_refseq=xref_refseq,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_vep_region(self, species, variants, blosum62=None, cadd=None,
                        gene_splicer=None, lo_f=None, max_ent_scan=None,
                        phenotypes=None, splice_ai=None, splice_region=None,
                        appris=None, canonical=None, ccds=None, db_nsfp=None,
                        dbsc_snv=None, distance=None, domains=None,
                        failed=None, hgvs=None, mane=None, merged=None,
                        mi_rna=None, minimal=None, numbers=None, protein=None,
                        refseq=None, shift_3prime=None, shift_genomic=None,
                        transcript_id=None, transcript_version=None, tsl=None,
                        uniprot=None, variant_class=None, vcf_string=None,
                        xref_refseq=None, norm_func=brc.default_pandas):
        """
        Fetch variant consequences for multiple regions. POST method. Queries:
        `vep/:species/region <https://rest.ensembl.org/documentation/info/vep_region_post>`_.
        Max post size: 200.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        variants : list
            Post option not defined in required/optional args.
        blosum62 : bool or NoneType, optional, default: NoneType
            Include BLOSUM62 amino acid conservation score (plugin details).
        cadd : bool or NoneType, optional, default: NoneType
            Include CADD (Combined Annotation Dependent Depletion)
            deleteriousness scores for single nucleotide variants. See license.
            (plugin details).
        gene_splicer : bool or NoneType, optional, default: NoneType
            Detects splice sites in genomic DNA (plugin details).
        lo_f : bool or NoneType, optional, default: NoneType
            LOFTEE identifies LoF (loss-of-function) variation. See README for
            more details.
        max_ent_scan : bool or NoneType, optional, default: NoneType
            Sequence motif and maximum entropy based splice site consensus
            predictions (plugin details).
        phenotypes : bool or NoneType, optional, default: NoneType
            Retrieves overlapping phenotype information (plugin details).
        splice_ai : bool or NoneType, optional, default: NoneType
            Retrieves pre-calculated annotations from SpliceAI a deep neural
            network, developed by Illumina, Inc that predicts splice junctions
            from an arbitrary pre-mRNA transcript sequence. Used for non-
            commercial purposes. (plugin details).
        splice_region : bool or NoneType, optional, default: NoneType
            More granular predictions of splicing effects (plugin details).
        appris : bool or NoneType, optional, default: NoneType
            Include APPRIS isoform annotation.
        canonical : bool or NoneType, optional, default: NoneType
            Include a flag indicating the canonical transcript for a gene.
        ccds : bool or NoneType, optional, default: NoneType
            Include CCDS transcript identifiers.
        db_nsfp : str or NoneType, optional, default: NoneType
            Include fields from dbNSFP, a database of pathogenicity predictions
            for missense variants. Multiple fields should be separated by
            commas. See dbNSFP README for field list. (plugin details). Example
            values: 'LRT_pred,MutationTaster_pred'.
        dbsc_snv : bool or NoneType, optional, default: NoneType
            Predictions for splicing variants from dbscSNV. (plugin details).
        distance : int or NoneType, optional, default: NoneType
            Change the distance to transcript for which VEP assigns upstream
            and downstream consequences.
        domains : bool or NoneType, optional, default: NoneType
            Include names of overlapping protein domains.
        failed : bool or NoneType, optional, default: NoneType
            When checking for co-located variants, by default variants flagged
            as failed by Ensembl's QC pipeline will be excluded. Set this flag
            to 1 to include such variants.
        hgvs : bool or NoneType, optional, default: NoneType
            Include HGVS nomenclature based on Ensembl stable identifiers.
        mane : bool or NoneType, optional, default: NoneType
            Include MANE Select annotations (GRCh38 only).
        merged : bool or NoneType, optional, default: NoneType
            Use merged Ensembl and RefSeq transcript set to report consequences
            (human only).
        mi_rna : bool or NoneType, optional, default: NoneType
            Determines where in the secondary structure of a miRNA a variant
            falls (plugin details).
        minimal : bool or NoneType, optional, default: NoneType
            Convert alleles to their most minimal representation before
            consequence calculation i.e. sequence that is identical between
            each pair of reference and alternate alleles is trimmed off from
            both ends, with coordinates adjusted accordingly. Note this may
            lead to discrepancies between input coordinates and coordinates
            reported by VEP relative to transcript sequences.
        numbers : bool or NoneType, optional, default: NoneType
            Include affected exon and intron positions within the transcript.
        protein : bool or NoneType, optional, default: NoneType
            Include Ensembl protein identifiers.
        refseq : bool or NoneType, optional, default: NoneType
            Use RefSeq transcript set to report consequences (human only).
        shift_3prime : bool or NoneType, optional, default: NoneType
            Shift transcript-overlapping variants as far as possible in the 3'
            direction before providing consequences.
        shift_genomic : bool or NoneType, optional, default: NoneType
            Shift all variants as far as possible in the 3' direction before
            providing consequences.
        transcript_id : str or NoneType, optional, default: NoneType
            Filter results by Transcript ID.
        transcript_version : bool or NoneType, optional, default: NoneType
            Add version numbers to Ensembl transcript identifiers.
        tsl : bool or NoneType, optional, default: NoneType
            Include transcript support level (TSL) annotation.
        uniprot : bool or NoneType, optional, default: NoneType
            Include best match accessions for translated protein products from
            three UniProt-related databases (SWISSPROT, TREMBL and UniParc).
        variant_class : bool or NoneType, optional, default: NoneType
            Output the Sequence Ontology variant class for the input variant.
        vcf_string : bool or NoneType, optional, default: NoneType
            Include alleles in VCF format.
        xref_refseq : bool or NoneType, optional, default: NoneType
            Include aligned RefSeq mRNA identifiers for transcript. NB:
            theRefSeq and Ensembl transcripts aligned in this way MAY NOT, AND
            FREQUENTLY WILL NOT, match exactly in sequence, exon structure and
            protein product.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        maximum_post_size = 200
        endpoint = 'vep/{species}/region'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            variants=variants,
            Blosum62=blosum62,
            CADD=cadd,
            GeneSplicer=gene_splicer,
            LoF=lo_f,
            MaxEntScan=max_ent_scan,
            Phenotypes=phenotypes,
            SpliceAI=splice_ai,
            SpliceRegion=splice_region,
            appris=appris,
            canonical=canonical,
            ccds=ccds,
            dbNSFP=db_nsfp,
            dbscSNV=dbsc_snv,
            distance=distance,
            domains=domains,
            failed=failed,
            hgvs=hgvs,
            mane=mane,
            merged=merged,
            miRNA=mi_rna,
            minimal=minimal,
            numbers=numbers,
            protein=protein,
            refseq=refseq,
            shift_3prime=shift_3prime,
            shift_genomic=shift_genomic,
            transcript_id=transcript_id,
            transcript_version=transcript_version,
            tsl=tsl,
            uniprot=uniprot,
            variant_class=variant_class,
            vcf_string=vcf_string,
            xref_refseq=xref_refseq,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Vep(brc._BaseRestClient, _Vep):
    """Perform Vep queries
    """
    pass



"""Perform queries against the ontology and taxonomy endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _OntologiesAndTaxonomy(object):
    """Perform OntologiesAndTaxonomy queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ontology_ancestors(self, query_id, ontology=None,
                               norm_func=brc.default_pandas):
        """
        Reconstruct the entire ancestry of a term from is_a and part_of
        relationships. GET method. Queries:
        `ontology/ancestors/:id <https://rest.ensembl.org/documentation/info/ontology_ancestors>`_.

        Parameters
        ----------
        query_id : str
            An ontology term identifier. Example values: 'GO:0005667'.
        ontology : str or NoneType, optional, default: NoneType
            Filter by ontology. Used to disambiguate terms which are shared
            between ontologies such as GO and EFO. Example values: 'GO'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ontology/ancestors/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            ontology=ontology,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ontology_ancestors_chart(self, query_id, ontology=None,
                                     norm_func=brc.default_pandas):
        """
        Reconstruct the entire ancestry of a term from is_a and part_of
        relationships. GET method. Queries:
        `ontology/ancestors/chart/:id <https://rest.ensembl.org/documentation/info/ontology_ancestors_chart>`_.

        Parameters
        ----------
        query_id : str
            An ontology term identifier. Example values: 'GO:0005667'.
        ontology : str or NoneType, optional, default: NoneType
            Filter by ontology. Used to disambiguate terms which are shared
            between ontologies such as GO and EFO. Example values: 'GO'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ontology/ancestors/chart/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            ontology=ontology,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ontology_descendants(self, query_id, closest_term=None,
                                 ontology=None, subset=None,
                                 zero_distance=None,
                                 norm_func=brc.default_pandas):
        """
        Find all the terms descended from a given term. By default searches are
        conducted within the namespace of the given identifier. GET method.
        Queries:
        `ontology/descendants/:id <https://rest.ensembl.org/documentation/info/ontology_descendants>`_.

        Parameters
        ----------
        query_id : str
            An ontology term identifier. Example values: 'GO:0005667'.
        closest_term : bool or NoneType, optional, default: NoneType
            If true return only the closest terms to the specified term.
        ontology : str or NoneType, optional, default: NoneType
            Filter by ontology. Used to disambiguate terms which are shared
            between ontologies such as GO and EFO. Example values: 'GO'.
        subset : str or NoneType, optional, default: NoneType
            Filter terms by the specified subset. Example values:
            'goslim_generic goslim_metagenomics'.
        zero_distance : bool or NoneType, optional, default: NoneType
            Return terms with a distance of 0.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ontology/descendants/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            closest_term=closest_term,
            ontology=ontology,
            subset=subset,
            zero_distance=zero_distance,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ontology_id(self, query_id, relation=None, simple=None,
                        norm_func=brc.default_pandas):
        """
        Search for an ontological term by its namespaced identifier. GET
        method. Queries:
        `ontology/id/:id <https://rest.ensembl.org/documentation/info/ontology_id>`_.

        Parameters
        ----------
        query_id : str
            An ontology term identifier. Example values: 'GO:0005667'.
        relation : str or NoneType, optional, default: NoneType
            The types of relationships to include in the output. Fetches all
            relations by default. Example values: 'is_a part_of'.
        simple : bool or NoneType, optional, default: NoneType
            If set the API will avoid the fetching of parent and child terms.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ontology/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            relation=relation,
            simple=simple,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_ontology_name(self, name, ontology=None, relation=None,
                          simple=None, norm_func=brc.default_pandas):
        """
        Search for a list of ontological terms by their name. GET method.
        Queries:
        `ontology/name/:name <https://rest.ensembl.org/documentation/info/ontology_name>`_.

        Parameters
        ----------
        name : str
            An ontology name. SQL wildcards are supported. Example values:
            'transcription factor complex'.
        ontology : str or NoneType, optional, default: NoneType
            Filter by ontology. Used to disambiguate terms which are shared
            between ontologies such as GO and EFO. Example values: 'GO'.
        relation : str or NoneType, optional, default: NoneType
            The types of relationships to include in the output. Fetches all
            relations by default. Example values: 'is_a part_of'.
        simple : bool or NoneType, optional, default: NoneType
            If set the API will avoid the fetching of parent and child terms.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'ontology/name/{name}'.format(
            name=name
        )

        return self.rest_query(
            method_type,
            endpoint,
            ontology=ontology,
            relation=relation,
            simple=simple,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_taxonomy_classification(self, query_id,
                                    norm_func=brc.default_pandas):
        """
        Return the taxonomic classification of a taxon node. GET method.
        Queries:
        `taxonomy/classification/:id <https://rest.ensembl.org/documentation/info/taxonomy_classification>`_.

        Parameters
        ----------
        query_id : str
            A taxon identifier. Can be a NCBI taxon id or a name. Example
            values: '9606 Homo sapiens'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'taxonomy/classification/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_taxonomy_id(self, query_id, simple=None,
                        norm_func=brc.default_pandas):
        """
        Search for a taxonomic term by its identifier or name. GET method.
        Queries:
        `taxonomy/id/:id <https://rest.ensembl.org/documentation/info/taxonomy_id>`_.

        Parameters
        ----------
        query_id : str
            A taxon identifier. Can be a NCBI taxon id or a name. Example
            values: '9606 Homo sapiens'.
        simple : bool or NoneType, optional, default: NoneType
            If set the API will avoid the fetching of parent and child terms.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'taxonomy/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            simple=simple,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_taxonomy_name(self, name, norm_func=brc.default_pandas):
        """
        Search for a taxonomic id by a non-scientific name. GET method. Queries:
        `taxonomy/name/:name <https://rest.ensembl.org/documentation/info/taxonomy_name>`_.

        Parameters
        ----------
        name : str
            A non-scientific species name. Can include SQL wildcards. Example
            values: 'Homo%25'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'taxonomy/name/{name}'.format(
            name=name
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class OntologiesAndTaxonomy(brc._BaseRestClient, _OntologiesAndTaxonomy):
    """Perform OntologiesAndTaxonomy queries
    """
    pass



"""Perform queries against the lookup endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc

try:
    # v2
    from pandas import json_normalize
except ImportError:
    # v 1.X
    from pandas.io.json import json_normalize

import pandas as pd
import itertools


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def NORM_POST_LOOKUP_ID_GENE(data, **kwargs):
    """
    When issuing a POST lookup ID query for a gene IDs, format returned
    data into a reasonable `pandas.DataFrame`. Not that missing IDs are not
    returned. Also, at present this will only work for 'gene' features

    Parameters
    ----------
    data : dict of dict
        The keys of the dictionary are GENE IDs and the values are dicts
        with nested data (if expand is True)
    **kwargs
        Arguments to the POST query that are required to process the
        different formats
    """
    try:
        # Determine the format
        data_format = kwargs['format']
    except KeyError:
        data_format = 'condensed'

    # Generate a list of dict and remove empty entries
    defined_data = [v for v in data.values() if v is not None]

    # These *_cols lists define the columns (as named after de-normalising)
    # and the column order we want in the final output. Not that as defined
    # here they are for the 'condensed' data_format
    exon_cols = ['trans_id', 'exon_id']
    gene_cols = ['gene_id', 'gene_version']
    trans_cols = ['trans_id', 'trans_version']
    prot_cols = ['trans_Translation_id']

    # The *_meta columns are the columns we want from the un-nested level
    # of the decoded JSON that we want to keep after de-normalising
    exon_meta = ['id']
    trans_meta = ['id', 'version']

    # The full data format has more info such as co-ordinates, as such we
    # adjust the columns we want to keep
    if data_format == 'full':
        gene_cols.extend(
            ['gene_display_name', 'gene_seq_region_name', 'gene_start',
             'gene_end', 'gene_strand', 'gene_biotype',
             'gene_description', 'gene_source']
        )
        trans_cols.extend(
            ['trans_display_name', 'trans_is_canonical',
             'trans_seq_region_name', 'trans_start', 'trans_end',
             'trans_strand', 'trans_biotype', 'trans_source']
        )
        prot_cols.extend(
            ['trans_Translation_start', 'trans_Translation_end',
             'trans_Translation_length']
        )
        exon_cols.extend(
            ['exon_version', 'exon_seq_region_name', 'exon_start',
             'exon_end', 'exon_strand']
        )
        trans_meta.extend(
            ['display_name', 'seq_region_name', 'start', 'end', 'strand',
             'biotype', 'description', 'source']
        )
    try:
        # If we have expanded results then we will have a lot lore nested
        # data on transcripts and exons and translations, this is expanded
        # using JSON normalise
        if kwargs['expand'] is True:
            # Extract a list of all the transcipts, this is done to get the
            # transcripts at the top level so I can do a separate
            # extraction of gene->transcript and transcript->exon and then
            # merge. I can't seem to noralize all in one go
            trans = itertools.chain.from_iterable(
                v['Transcript'] for v in defined_data
            )
            # Normalise exons/transcript and extract columns
            exons = json_normalize(
                trans,
                meta_prefix='trans_',
                record_prefix='exon_',
                meta=exon_meta,
                record_path=['Exon']
            )[exon_cols]
            trans = json_normalize(
                defined_data,
                meta_prefix='gene_',
                record_prefix='trans_',
                meta=trans_meta,
                record_path="Transcript",
                max_level=1,
                errors='ignore',
                sep='_'
            )[gene_cols + trans_cols + prot_cols]
            trans.columns = map(str.lower, trans.columns)
            return trans.merge(
                exons, left_on='trans_id', right_on='trans_id'
            ).set_index(['gene_id', 'trans_id', 'exon_id'])
    except KeyError:
        pass

    # Here we are processing unexpanded data but we still want to have some
    # sensible column naming and ordering
    df = pd.DataFrame(defined_data)
    df.columns = pd.Index(['gene_{0}'.format(i) for i in df.columns])
    return df[gene_cols].set_index('gene_id')


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Lookup(object):
    """Perform Lookup queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_lookup_id(self, query_id, db_type=None, expand=None,
                      data_format=None, phenotypes=None, species=None,
                      utr=None, norm_func=brc.default_pandas):
        """
        Find the species and database for a single identifier e.g. gene,
        transcript, protein. GET method. Queries:
        `lookup/id/:id <https://rest.ensembl.org/documentation/info/lookup>`_.

        Parameters
        ----------
        query_id : str
            An Ensembl stable ID. Example values: 'ENSG00000157764'.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        expand : bool or NoneType, optional, default: NoneType
            Expands the search to include any connected features. e.g. If the
            object is a gene, its transcripts, translations and exons will be
            returned as well.
        data_format : str or NoneType, optional, default: NoneType
            Specify the formats to emit from this endpoint. Allowed values:
            full,condensed
        phenotypes : bool or NoneType, optional, default: NoneType
            Include phenotypes. Only available for gene objects.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Example values: 'homo_sapiens human'.
        utr : bool or NoneType, optional, default: NoneType
            Include 5' and 3' UTR features. Only available if the expand option
            is used.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        result_format = 'dict'
        endpoint = 'lookup/id/{id}'.format(
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            db_type=db_type,
            expand=expand,
            format=data_format,
            phenotypes=phenotypes,
            species=species,
            utr=utr,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_lookup_id(self, ids, db_type=None, expand=None, data_format=None,
                       object_type=None, species=None,
                       norm_func=NORM_POST_LOOKUP_ID_GENE):
        """
        Find the species and database for several identifiers. IDs that are not
        found are returned with no data. POST method. Queries:
        `lookup/id <https://rest.ensembl.org/documentation/info/lookup_post>`_.
        Max post size: 1000.

        Parameters
        ----------
        ids : list
            Post option not defined in required/optional args.
        db_type : str or NoneType, optional, default: NoneType
            Restrict the search to a database other than the default. Useful if
            you need to use a DB other than core. Example values: 'core
            otherfeatures'.
        expand : bool or NoneType, optional, default: NoneType
            Expands the search to include any connected features. e.g. If the
            object is a gene, its transcripts, translations and exons will be
            returned as well.
        data_format : str or NoneType, optional, default: NoneType
            Specify the formats to emit from this endpoint. Allowed values:
            full,condensed
        object_type : str or NoneType, optional, default: NoneType
            Filter by feature type. Example values: 'gene','transcript'.
        species : str or NoneType, optional, default: NoneType
            Species name/alias. Causes problems if the species doesn't match
            the identifiers in the POST body. Example values: 'homo_sapiens
            human'.
        norm_func : obj:`func`, optional, default: NORM_POST_LOOKUP_ID_GENE
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        result_format = 'dict'
        maximum_post_size = 1000
        endpoint = 'lookup/id'

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            ids=ids,
            db_type=db_type,
            expand=expand,
            format=data_format,
            object_type=object_type,
            species=species,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_lookup_symbol(self, species, symbol, expand=None,
                          data_format=None, norm_func=brc.default_pandas):
        """
        Find the species and database for a symbol in a linked external
        database. GET method. Queries:
        `lookup/symbol/:species/:symbol <https://rest.ensembl.org/documentation/info/symbol_lookup>`_.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        symbol : str
            A name or symbol from an annotation source has been linked to a
            genetic feature. Example values: 'BRCA2'.
        expand : bool or NoneType, optional, default: NoneType
            Expands the search to include any connected features. e.g. If the
            object is a gene, its transcripts, translations and exons will be
            returned as well.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'lookup/symbol/{species}/{symbol}'.format(
            species=species,
            symbol=symbol
        )

        return self.rest_query(
            method_type,
            endpoint,
            expand=expand,
            format=data_format,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_lookup_symbol(self, species, symbols, expand=None,
                           data_format=None, norm_func=brc.default_pandas):
        """
        Find the species and database for a set of symbols in a linked external
        database. Unknown symbols are omitted from the response. POST method.
        Queries:
        `lookup/symbol/:species/:symbol <https://rest.ensembl.org/documentation/info/symbol_post>`_.
        Max post size: 1000.

        Parameters
        ----------
        species : str
            Species name/alias for the whole batch of symbols. Example values:
            'homo_sapiens human'.
        symbols : list
            Post option not defined in required/optional args.
        expand : bool or NoneType, optional, default: NoneType
            Expands the search to include any connected features. e.g. If the
            object is a gene, its transcripts, translations and exons will be
            returned as well.
        data_format : str or NoneType, optional, default: NoneType
            Specify the layout of the response. Allowed values: full,condensed
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        raise NotImplementedError(
            'I think this method call is broken actually in the rest API'
        )
        method_type = 'POST'
        maximum_post_size = 1000
        endpoint = 'lookup/symbol/{species}/{symbol}'.format(
            species=species,
            symbol=symbols
        )

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            expand=expand,
            format=data_format,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Lookup(brc._BaseRestClient, _Lookup):
    """Perform Lookup queries
    """
    pass



"""Perform queries against the variation endpoints
"""
from ensembl_rest_client.base import base_rest_client as brc


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _Variation(object):
    """Perform Variation queries
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_variant_recoder(self, query_id, species, fields=None,
                            norm_func=brc.default_pandas):
        """
        Translate a variant identifier, HGVS notation or genomic SPDI notation
        to all possible variant IDs, HGVS and genomic SPDI. GET method. Queries:
        `variant_recoder/:species/:id <https://rest.ensembl.org/documentation/info/variant_recoder>`_.

        Parameters
        ----------
        query_id : str
            Variant ID, HGVS notation or genomic SPDI notation. Example values:
            'rs56116432 AGT:c.803T>C NC_000023.11:284252:C:G'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        fields : str or NoneType, optional, default: NoneType
            Comma-separated list of identifiers/notations to include from the
            following types: id (variant ID), hgvsg (HGVS genomic), hgvsc (HGVS
            coding), hgvsp (HGVS protein), spdi (SPDI genomic), vcf_string (VCF
            format).
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'variant_recoder/{species}/{id}'.format(
            species=species,
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            fields=fields,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_variant_recoder(self, species, ids, fields=None,
                             norm_func=brc.default_pandas):
        """
        Translate a list of variant identifiers, HGVS notations or genomic SPDI
        notations to all possible variant IDs, HGVS and genomic SPDI. POST
        method. Queries:
        `variant_recoder/:species <https://rest.ensembl.org/documentation/info/variant_recoder_post>`_.
        Max post size: 200.

        Parameters
        ----------
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        ids : list
            Post option not defined in required/optional args.
        fields : str or NoneType, optional, default: NoneType
            Comma-separated list of identifiers/notations to include from the
            following types: id (variant ID), hgvsg (HGVS genomic), hgvsc (HGVS
            coding), hgvsp (HGVS protein), spdi (SPDI genomic), vcf_string (VCF
            format).
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        maximum_post_size = 200
        endpoint = 'variant_recoder/{species}'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            ids=ids,
            fields=fields,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_variation(self, query_id, species, genotypes=None,
                      genotyping_chips=None, phenotypes=None, pops=None,
                      population_genotypes=None, norm_func=brc.default_pandas):
        """
        Uses a variant identifier (e.g. rsID) to return the variation features
        including optional genotype, phenotype and population data. GET method.
        Queries:
        `variation/:species/:id <https://rest.ensembl.org/documentation/info/variation_id>`_.

        Parameters
        ----------
        query_id : str
            Variant id. Example values: 'rs56116432'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        genotypes : bool or NoneType, optional, default: NoneType
            Include individual genotypes.
        genotyping_chips : bool or NoneType, optional, default: NoneType
            Include genotyping chips information.
        phenotypes : bool or NoneType, optional, default: NoneType
            Include phenotypes.
        pops : bool or NoneType, optional, default: NoneType
            Include population allele frequencies.
        population_genotypes : bool or NoneType, optional, default: NoneType
            Include population genotype frequencies.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'variation/{species}/{id}'.format(
            species=species,
            id=query_id
        )

        return self.rest_query(
            method_type,
            endpoint,
            genotypes=genotypes,
            genotyping_chips=genotyping_chips,
            phenotypes=phenotypes,
            pops=pops,
            population_genotypes=population_genotypes,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_variation_pmcid(self, pmcid, species,
                            norm_func=brc.default_pandas):
        """
        Fetch variants by publication using PubMed Central reference number
        (PMCID). GET method. Queries:
        `variation/:species/pmcid/:pmcid <https://rest.ensembl.org/documentation/info/variation_pmcid_get>`_.

        Parameters
        ----------
        pmcid : str
            PubMed Central reference number (PMCID). Example values:
            'PMC5002951'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'variation/{species}/pmcid/{pmcid}'.format(
            species=species,
            pmcid=pmcid
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def get_variation_pmid(self, pmid, species, norm_func=brc.default_pandas):
        """
        Fetch variants by publication using PubMed reference number (PMID). GET
        method. Queries:
        `variation/:species/pmid/:pmid <https://rest.ensembl.org/documentation/info/variation_pmid_get>`_.

        Parameters
        ----------
        pmid : str
            PubMed reference number (PMID). Example values: '26318936'.
        species : str
            Species name/alias. Example values: 'homo_sapiens human'.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'GET'
        endpoint = 'variation/{species}/pmid/{pmid}'.format(
            species=species,
            pmid=pmid
        )

        return self.rest_query(
            method_type,
            endpoint,
            norm_func=norm_func
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def post_variation(self, species, ids, genotypes=None, phenotypes=None,
                       pops=None, population_genotypes=None,
                       norm_func=brc.default_pandas):
        """
        Uses a list of variant identifiers (e.g. rsID) to return the variation
        features including optional genotype, phenotype and population data.
        POST method. Queries:
        `variation/:species/ <https://rest.ensembl.org/documentation/info/variation_post>`_.
        Max post size: 200.

        Parameters
        ----------
        species : str
            Species name/alias for the whole batch of symbols. Example values:
            'homo_sapiens human'.
        ids : list
            Post option not defined in required/optional args.
        genotypes : bool or NoneType, optional, default: NoneType
            Include individual genotypes.
        phenotypes : bool or NoneType, optional, default: NoneType
            Include phenotypes.
        pops : bool or NoneType, optional, default: NoneType
            Include population allele frequencies.
        population_genotypes : bool or NoneType, optional, default: NoneType
            Include population genotype frequencies.
        norm_func : obj:`func`, optional, default: default_pandas
            The normalisation function to use when obj.pandas is True

        Returns
        -------
        data : list or dict or :obj:`pandas.DataFrame`
            The rest query data
        """
        method_type = 'POST'
        maximum_post_size = 200
        endpoint = 'variation/{species}/'.format(
            species=species
        )

        return self.rest_query(
            method_type,
            endpoint,
            maximum_post_size=maximum_post_size,
            ids=ids,
            genotypes=genotypes,
            phenotypes=phenotypes,
            pops=pops,
            population_genotypes=population_genotypes,
            norm_func=norm_func
        )


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class Variation(brc._BaseRestClient, _Variation):
    """Perform Variation queries
    """
    pass
